from fastapi import FastAPI, APIRouter, Depends, Request, Path
from fastapi.encoders import SetIntStr, DictIntStrAny
from fastapi.routing import APIRoute
from fastapi.datastructures import Default
from pydantic import BaseConfig, BaseModel, create_model
from pydantic.fields import ModelField
from starlette.routing import BaseRoute
from starlette.responses import Response, JSONResponse
from starlette.types import ASGIApp
from typing import Optional, List, Sequence, Type, Any, Callable, Union, Dict
from sqlalchemy.orm import decl_api
from sqlalchemy.ext.asyncio import AsyncSession
from types import FunctionType
from datetime import datetime

from .dependencies.utils import BaseCRUD
from .dependencies.utility import CommonQueryGetter

class SimpleEndpoint:
    def __init__(self, 
            path: str = "",
            enable: bool = True,
            response_model: Optional[Type[Any]] = None,
            status_code: Optional[int] = None,
            tags: Optional[List[str]] = None,
            dependencies: Optional[Sequence[Depends]] = None,
            summary: Optional[str] = None,
            description: Optional[str] = None,
            response_description: str = "Successful Response",
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            deprecated: Optional[bool] = None,
            operation_id: Optional[str] = None,
            response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
            response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
            response_model_by_alias: bool = True,
            response_model_exclude_unset: bool = False,
            response_model_exclude_defaults: bool = False,
            response_model_exclude_none: bool = False,
            include_in_schema: bool = True,
            response_class: Type[Response] = Default(JSONResponse),
            name: Optional[str] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            openapi_extra: Optional[Dict[str, Any]] = None,
            pydantic_model: Optional[BaseModel] = None
        ):
        self.enable = enable
        self.path = path
        self.response_model = response_model
        self.status_code = status_code
        self.tags = tags
        self.dependencies = dependencies
        self.summary = summary
        self.description = description
        self.response_description = response_description
        self.responses = responses
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.response_model_include = response_model_include
        self.response_model_exclude = response_model_exclude
        self.response_model_by_alias = response_model_by_alias
        self.response_model_exclude_unset = response_model_exclude_unset
        self.response_model_exclude_defaults = response_model_exclude_defaults
        self.response_model_exclude_none = response_model_exclude_none
        self.include_in_schema = include_in_schema
        self.response_class = response_class
        self.name = name
        self.callbacks = callbacks
        self.openapi_extra = openapi_extra
        self.modelPydantic = pydantic_model
    
    def get_endpoint_kwargs(self):
        params = vars(self).copy()
        params.pop("enable")
        params.pop("modelPydantic")
        return params    


class SimpleRouter(APIRouter):
    """
    Simple Router Class to define a router for SimpleCRUDGenerator
    """
    def __init__(
            self,
            classModel: decl_api.DeclarativeMeta,
            *,
            prefix: str = "",
            tags: Optional[List[str]] = None,
            dependencies: Optional[Sequence[Depends]] = None,
            default_response_class: Type[Response] = Default(JSONResponse),
            responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
            callbacks: Optional[List[BaseRoute]] = None,
            routes: Optional[List[BaseRoute]] = None,
            redirect_slashes: bool = True,
            default: Optional[ASGIApp] = None,
            dependency_overrides_provider: Optional[Any] = None,
            route_class: Type[APIRoute] = APIRoute,
            on_startup: Optional[Sequence[Callable[[], Any]]] = None,
            on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
            deprecated: Optional[bool] = None,
            include_in_schema: bool = True,
            crud_create: Union[SimpleEndpoint, bool, None] = True,
            crud_read: Union[SimpleEndpoint, bool, None] = True,
            crud_update: Union[SimpleEndpoint, bool, None] = True,
            crud_delete: Union[SimpleEndpoint, bool, None] = True,
            disable_simple_crud: bool = False
        ):
        self.classModel = classModel
        self.tablename = classModel.__tablename__
        self.modelPydanticforCreate = self.generate_pydantic_model(modelName=self.tablename+"PydanticSimpleCreate")
        self.modelPydanticforUpdate = self.generate_pydantic_model(modelName=self.tablename+"PydanticSimpleUpdate", excluded_attributes=["id"])
        self.crud = BaseCRUD(classModel)
        if not tags: tags = [self.tablename]
        if not prefix: prefix = f"/{self.tablename}"
        super().__init__(
                prefix=prefix,
                tags=tags,
                dependencies=dependencies,
                default_response_class=default_response_class,
                responses=responses,
                callbacks=callbacks,
                routes=routes,
                redirect_slashes=redirect_slashes,
                default=default,
                dependency_overrides_provider=dependency_overrides_provider,
                route_class=route_class,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                deprecated=deprecated,
                include_in_schema=include_in_schema)
        if disable_simple_crud:
            crud_create = None
            crud_read = None
            crud_update = None
            crud_delete = None
        if type(crud_create) == SimpleEndpoint:
            self.crud_create = crud_create
        else:
            if crud_create:
                self.crud_create = SimpleEndpoint(enable=True)
            else:
                self.crud_create = SimpleEndpoint(enable=False)
        if type(crud_read) == SimpleEndpoint:
            self.crud_read = crud_read
        else:
            if crud_read:
                self.crud_read = SimpleEndpoint(enable=True)
            else:
                self.crud_read = SimpleEndpoint(enable=False)
        if type(crud_update) == SimpleEndpoint:
            self.crud_update = crud_update
        else:
            if crud_update:
                self.crud_update = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.crud_update = SimpleEndpoint(enable=False)
        if type(crud_delete) == SimpleEndpoint:
            self.crud_delete = crud_delete
        else:
            if crud_delete:
                self.crud_delete = SimpleEndpoint(path="/{id}", enable=True)
            else:
                self.crud_delete = SimpleEndpoint(enable=False)
        self.__get_session = None
    
    def set_the_get_session(self, method: FunctionType):
        self.__get_session = method

    def generate_pydantic_model(self, modelName: str = "", excluded_attributes=[]):
        annots = self.get_annotation()
        for ex_at in excluded_attributes:
            if ex_at in annots: annots.pop(ex_at)
        if not modelName: modelName = self.tablename+"Pydantic"
        ModelPydantic = create_model(modelName, **annots)
        return ModelPydantic
    
    def get_annotation(self):
        try:
            classModel = self.classModel
            fields = [i for i in vars(classModel) if "_" not in [i[0], i[-1]]]
            keyValuePair = {}
            for f in fields:
                if "comparator" in classModel.__dict__[f].__dict__:
                    if "column" in str(classModel.__dict__[f].__dict__["comparator"]).lower():
                        comp = classModel.__dict__[f].__dict__["comparator"]
                        if "VARCHAR" in str(comp.type):
                            if "enum_class" in comp.type.__dict__:
                                if comp.nullable:
                                    keyValuePair[f] = (Optional[comp.type.enum_class], None)
                                else:
                                    keyValuePair[f] = (comp.type.enum_class, None)                                    
                            else:
                                if comp.nullable:
                                    keyValuePair[f] = (Optional[str], None)
                                else:
                                    keyValuePair[f] = (str, None)
                        elif "TEXT" in str(comp.type):
                            if comp.nullable:
                                keyValuePair[f] = (Optional[str], None)
                            else:
                                keyValuePair[f] = (str, None)
                        elif "BOOLEAN" in str(comp.type):
                            if comp.nullable:
                                keyValuePair[f] = (Optional[bool], None)
                            else:
                                keyValuePair[f] = (bool, None)
                        elif "INTEGER" in str(comp.type):
                            if comp.nullable:
                                keyValuePair[f] = (Optional[int], None)
                            else:
                                keyValuePair[f] = (int, None)
                        elif "FLOAT" in str(comp.type):
                            if comp.nullable:
                                keyValuePair[f] = (Optional[float], None)
                            else:
                                keyValuePair[f] = (float, None)
                        elif "DATETIME" in str(comp.type):
                            if comp.nullable:
                                keyValuePair[f] = (Optional[datetime], None)
                            else:
                                keyValuePair[f] = (datetime, None)
            return keyValuePair
        except:
            return {}

    def _setup_crud(self):
        if self.crud_create.enable:
            kargs = self.crud_create.get_endpoint_kwargs()
            if not kargs["name"]:
                kargs["name"] = "create "+self.tablename
            if self.crud_create.modelPydantic:
                modelPydantic_ = self.crud_create.modelPydantic
            else:
                modelPydantic_ = self.modelPydanticforCreate
            @self.post(**kargs)
            async def base_post(
                    request: Request,
                    modelPydantic: modelPydantic_,
                    session: AsyncSession = Depends(self.__get_session)
                ):
                return await self.crud.create(modelPydantic, session)
        
        if self.crud_read.enable:
            kargs = self.crud_read.get_endpoint_kwargs()
            if not kargs["name"]:
                kargs["name"] = "read "+self.tablename
            @self.get(**kargs)
            async def base_get(
                    request: Request,
                    getParams = Depends(CommonQueryGetter),
                    session: AsyncSession = Depends(self.__get_session)
                ):
                return await self.crud.read(getParams, session)

        if self.crud_update.enable:
            kargs = self.crud_update.get_endpoint_kwargs()
            if not kargs["name"]:
                kargs["name"] = "update "+self.tablename
            if self.crud_update.modelPydantic:
                modelPydantic_ = self.crud_update.modelPydantic
            else:
                modelPydantic_ = self.modelPydanticforUpdate
            @self.put(**kargs)
            async def base_put(
                    request: Request,
                    modelPydantic: modelPydantic_,
                    id: int = Path(...,min=1),
                    session: AsyncSession = Depends(self.__get_session)
                ):
                return await self.crud.update(modelPydantic, id, session)

        if self.crud_delete.enable:
            kargs = self.crud_delete.get_endpoint_kwargs()
            if not kargs["name"]:
                kargs["name"] = "delete "+self.tablename
            @self.delete(**kargs)
            async def base_delete(
                    request: Request,
                    id: int = Path(...,min=1),
                    session: AsyncSession = Depends(self.__get_session)
                ):
                return await self.crud.delete(id, session)


class RouterMap():
    """
    Router Mapping Class to define and simplify your router simple CRUD
    """
    @classmethod
    def _get_variable(cls):
        return [cls.__dict__[v] for v in vars(cls) if type(cls.__dict__[v]) == SimpleRouter]
    
    @classmethod
    def _get_key_value(cls):
        return {v: cls.__dict__[v] for v in vars(cls) if type(cls.__dict__[v]) == SimpleRouter}

    @classmethod
    def _collect_simple_router(cls):
        allRouters = {}
        for c in cls.__subclasses__():
            for v in c._get_key_value().values():
                allRouters[v.tablename] = v
            if c.__subclasses__():
                allRouters.update(c._collect_simple_router())                
        return allRouters
    
    @staticmethod
    def create_router_map_from_base(base: decl_api.DeclarativeMeta, base_prefix: str = ""):
        """
        Create All CRUD Automatically from SQLAlchemy declared base
        """
        class AutoMap(RouterMap): pass
        for c in base.__subclasses__():
            setattr(
                AutoMap,
                c.__tablename__,                
                SimpleRouter(c, prefix=base_prefix+"/"+c.__tablename__, tags=[c.__tablename__])
                )
        return AutoMap
        

class SimpleCRUDGenerator():
    """
    Create CRUD Operation in a simple way.
    
    params:
    - application -> FastAPI Application
    - session_getter -> method to get the sqlalchemy session
    - autogenerate (bool) -> once instantiated, the base crud endpoints will be created
    """
    def __init__(self, application: Optional[FastAPI] = None, session_getter: Optional[FunctionType] = None, autogenerate: bool = True):
        self.app = application
        self.allRouters = RouterMap._collect_simple_router()
        self.session_getter = session_getter
        if not all([application, session_getter]): autogenerate = False
        if autogenerate: self.generate_router()
    
    def set_application(self, application: FastAPI):
        """
        Set your FastAPI object application
        """
        self.app = application
    
    def set_session_getter(self, session_getter: FunctionType):
        """
        Set your session getter for SQLAlchemy
        """
        self.session_getter = session_getter

    def add_simple_router(self, simple_router: SimpleRouter):
        """
        Add your fastapi_simple_crud.SimpleRouter object to be generated outside of the router map
        """
        self.allRouters[simple_router.tablename] = simple_router

    def generate_router(self, application: Optional[FastAPI] = None, session_getter: Optional[FunctionType] = None):
        """
        Generate defined router map
        """
        if application: self.set_application(application)
        if session_getter: self.set_session_getter(session_getter)
        if all([self.app, self.session_getter]):
            for tag in sorted(self.allRouters):
                router = self.allRouters[tag]
                router.set_the_get_session(self.session_getter)
                router._setup_crud()
                self.app.include_router(router)


from datetime import datetime
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.orm import load_only, decl_api
from sqlalchemy import asc, desc, func
from sqlalchemy.sql.selectable import Select
from typing import Optional, Union

from .status import StatusResponse
from .utility import CommonQueryGetter
from .log import logger


def create_status(code, message_server, message_client=None):
    if message_client is None:
        message_client = message_server
    return {"code": code, "message": message_server}


def create_response(data={}, meta={}, status={}):
    return {
        "data": data,
        "meta": meta,
        "status": status,
    }

async def update_data(session, tableModel, dataObject, newDataPydantic):
    updates = {}
    fields = [i for i in vars(tableModel) if "_" not in [i[0], i[-1]]]
    for key in fields:
        if key in newDataPydantic.__dict__:
            updates[key] = newDataPydantic.__dict__[key]
    for key, value in updates.items():
        setattr(dataObject, key, value)
    session.add(dataObject)
    await session.commit()
    await session.refresh(dataObject)

class StatusCreator(StatusResponse):
    """
    to add more status, edit dependencies/status.py in StatusResponse class
    """

    def __init__(self):
        self._applyStatusAsAttribute()

    def _applyStatusAsAttribute(self):
        allAttr = [i for i in vars(StatusResponse) if not "_" in [i[0], i[-1]]]
        for a in allAttr:

            def fetcher(attrName):
                def response(message=str(attrName).replace("_", " ")):
                    code = StatusResponse.__dict__[attrName]
                    return create_status(code, str(message))

                return response

            self.__dict__[a] = fetcher(a)


status = StatusCreator()
"""
- The default message is the status name
- You can add your custom message by inserting your message to the function
    - ex: status.success("great")
- To add more status, edit 'StatusResponse' class in dependencies/status.py
"""


def get_field(classModel):
    try:
        fields = [i for i in vars(classModel) if "_" not in [i[0], i[-1]]]
        columnFields = []
        for f in fields:
            if "comparator" in classModel.__dict__[f].__dict__:
                if (
                    "column"
                    in str(classModel.__dict__[f].__dict__["comparator"]).lower()
                ):
                    columnFields.append(f)
        return columnFields
    except:
        return []

def validate_field(modelField, targetField):
    if type(modelField) in [list, tuple, dict]:
        availableField = modelField
    else:
        availableField = get_field(modelField)
    return [i for i in targetField if i in availableField]

class SimpleCRUDBaseDeclaration:
    base = None
    
    @staticmethod
    def set_base(base: decl_api.DeclarativeMeta):
        if type(base) == decl_api.DeclarativeMeta:
            SimpleCRUDBaseDeclaration.base = base
        else:
            raise BaseException("Invalid Base")

def set_declarative_base(base: decl_api.DeclarativeMeta):
    return SimpleCRUDBaseDeclaration.set_base(base)

class Selector:
    """
    Selector Class, modifiying the sqlalchemy select
    """

    def __init__(self, *columns, **columnsKeyPair):
        """
        add your customer key by adding the key-value pair
            ex: Selector(uid = User.id, name = Profile.name)

        or

        define it automatically by just giving the column name
            ex: Selector(User.id, Profile.name)

        (this output column name will be user_id, profile_name)
        """
        if not SimpleCRUDBaseDeclaration.base:
            msg = """
            SQLAlchemy Declarative Base is not defined.

            please define it by using 'set_declarative_base':

                from fastapi_simple_crud.dependencies import set_declarative_base

                set_declarative_base(Base)
            """
            raise BaseException(msg)
        self.baseClass = SimpleCRUDBaseDeclaration.base.__subclasses__()
        if columnsKeyPair:
            self.columns = list(columnsKeyPair.values())
            self.columnsKeyPair = columnsKeyPair
        else:
            self.columns = columns
            self.columnsKeyPair = dict(zip(self.get_keys(), list(columns)))
        self.keys = list(self.columnsKeyPair.keys())

    @property
    def query(self):
        return select(*self.columns)

    def get_keys(self):
        keys = []
        for c in self.columns:
            className, key = str(c).split(".")
            className = self.check_belonging(className)
            keys.append(className + "_" + key)
        return keys

    def check_belonging(self, key):
        for c in self.baseClass:
            cKey = str(c).split("'")[1].split(".")[-1]
            if key == cKey:
                return c.__tablename__
        return key

    async def execute(self, session: AsyncSession, query: Select):
        data = []
        query = await session.execute(query)
        for q in query:
            data.append(dict(zip(self.keys, list(q))))
        return data


class QueryManager:
    def __init__(self, classModel):
        self.classModel = classModel

    @property
    def fields(self) -> list:
        return get_field(self.classModel)

    @property
    def rawQuery(self) -> Select:
        return select(self.classModel)

    def validate_fields(self, fields: list) -> list:
        if fields:
            return [i for i in fields if i in self.fields]
        return []


def QueryPaginator(
    getParams: CommonQueryGetter, _obj: Union[decl_api.DeclarativeMeta, Selector]
):
    """
    Query Paginator generator for single or multiple table
    """
    if type(_obj) == Selector:
        return QueryPaginatorMultiple(getParams, _obj)
    if _obj.__class__ == decl_api.DeclarativeMeta:
        if len(_obj.__mro__) == 3:
            return QueryPaginatorSingle(getParams, _obj)
    raise BaseException("Error Paginator")


class QueryPaginatorSingle(QueryManager):
    """
    Query Paginator for common use of paginating functions
    """

    def __init__(self, getParams: CommonQueryGetter, classModel):
        QueryManager.__init__(self, classModel)
        self.getParams = getParams
        if getParams.fields:
            self.filterFields = self.validate_fields(getParams.fields.split(","))
        else:
            self.filterFields = self.fields

    def filter(self, query: Select, fields: Optional[list] = None) -> Select:
        """
        filter the desired fields from the query (it will remains the primary key)
        """
        if not fields:
            fields = self.filterFields
        return query.options(load_only(*fields))

    def filter_primary_key(self, data: list) -> list:
        """
        primary key filter
        """
        newData = []
        for d in data:
            newPair = {}
            for key in d.__dict__:
                if key in self.filterFields:
                    newPair[key] = d.__dict__[key]
            newData.append(newPair)
        return newData

    def sort(self, query: Select, sortBy: str, sortType: str) -> Select:
        """
        sort your query by defining both the sortBy 'fields' and sortType 'asc'/'desc' parameters
        """
        if sortBy in self.fields:
            merchantOrder = self.classModel.__dict__[sortBy]
            for iSortType, orderMethod in [["asc", asc], ["desc", desc]]:
                if sortType == iSortType:
                    return query.order_by(orderMethod(merchantOrder))
        return query

    def paginate(self, query: Select, pageNo: int, limitPerPage: int) -> Select:
        """
        paginate query by defining page number and page limit
        """
        query = query.offset((pageNo - 1) * limitPerPage)
        query = query.limit(limitPerPage)
        return query

    async def execute_pagination(self, session: AsyncSession, query: Select) -> dict:
        """
        sort, filter and paginate the query with executed query outputs
        """
        query = self.filter(query)
        countAfterFilter = await session.execute(
            select(func.count()).select_from(query.subquery())
        )
        countAfterFilter = countAfterFilter.scalars().one()
        query = self.sort(query, self.getParams.sortBy, self.getParams.sortType)
        query = self.paginate(query, self.getParams.page, self.getParams.limit)
        query = await session.execute(query)
        datas = query.scalars().all()
        datas = self.filter_primary_key(datas)
        return create_response(
            data={"list": datas},
            meta={
                "page": self.getParams.page,
                "length": self.getParams.limit,
                "total": countAfterFilter,
            },
            status=status.success(),
        )


class QueryPaginatorMultiple:
    def __init__(self, getParams: CommonQueryGetter, selector: Selector):
        self.rawQuery = selector.query
        self.getParams = getParams
        self.selector = selector
        if getParams.fields:
            self.fields = validate_field(
                self.selector.keys, getParams.fields.split(",")
            )
        else:
            self.fields = self.selector.keys

    def filter(self, data: list, fields: Optional[list] = None) -> Select:
        """
        filter the desired fields from the query (it will remains the primary key)
        """
        if not fields:
            fields = self.fields
        newData = []
        for d in data:
            newPair = {}
            for k in d:
                if k in fields:
                    newPair[k] = d[k]
            newData.append(newPair)
        return newData

    def sort(self, query: Select, sortBy: str, sortType: str) -> Select:
        """
        sort your query by defining both the sortBy 'fields' and sortType 'asc'/'desc' parameters
        """
        if sortBy in self.fields:
            merchantOrder = self.selector.columnsKeyPair[sortBy]
            for iSortType, orderMethod in [["asc", asc], ["desc", desc]]:
                if sortType == iSortType:
                    return query.order_by(orderMethod(merchantOrder))
        return query

    def paginate(self, query: Select, pageNo: int, limitPerPage: int) -> Select:
        """
        paginate query by defining page number and page limit
        """
        query = query.offset((pageNo - 1) * limitPerPage)
        query = query.limit(limitPerPage)
        return query

    async def execute_pagination(self, session: AsyncSession, query: Select):
        countAfterFilter = await session.execute(
            select(func.count()).select_from(query.subquery())
        )
        countAfterFilter = countAfterFilter.scalars().one()
        query = self.sort(query, self.getParams.sortBy, self.getParams.sortType)
        query = self.paginate(query, self.getParams.page, self.getParams.limit)
        datas = await self.selector.execute(session, query)
        datas = self.filter(datas)
        return create_response(
            data={"list": datas},
            meta={
                "page": self.getParams.page,
                "length": self.getParams.limit,
                "total": countAfterFilter,
            },
            status=status.success(),
        )


class BaseWhereClause:
    def __init__(self, classModel, *whereExpression, **whereClause):
        self.classModel = classModel
        self.we = whereExpression
        self.wc = whereClause

    def applyWhereObject(self, query: Select, whereClauseDict, **WhereClauseKwargs):
        WhereClauseKwargs.update(whereClauseDict)
        self.wc.update(WhereClauseKwargs)
        for w in self.we:
            query = query.where(w)
        for key in self.wc:
            query = query.where(self.classModel.__dict__[key] == self.wc[key])
        return query


class BaseCRUD:
    def __init__(self, classModel):
        self.classModel = classModel

    def where(self, *whereExpression, **whereClause):
        return BaseWhereClause(self.classModel, *whereExpression, **whereClause)

    async def read(self, getParams: CommonQueryGetter, session: AsyncSession):
        try:
            paginator = QueryPaginator(getParams, self.classModel)
            query = paginator.rawQuery
            res = await paginator.execute_pagination(session, query)
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def create(self, pydanticModel, session: AsyncSession):
        try:
            newData = self.classModel(**pydanticModel.dict())
            session.add(newData)
            await session.commit()
            await session.refresh(newData)
            res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def update(
        self,
        pydanticModel,
        id: Optional[int],
        session: AsyncSession,
        whereClauseObject: Optional[BaseWhereClause] = None,
        **whereClause
    ):
        try:
            query = select(self.classModel)
            if id != None:
                query = query.where(self.classModel.id == id)
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            data = await session.execute(query)
            data = data.scalars().first()
            if not data:
                res = create_response(status=status.data_not_updated())
            else:
                await update_data(session, self.classModel, data, pydanticModel)
                res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res

    async def delete(
        self,
        id: Optional[int],
        session: AsyncSession,
        whereClauseObject: Optional[BaseWhereClause] = None,
        **whereClause
    ):
        try:
            query = select(self.classModel)
            if id != None:
                query = query.where(self.classModel.id == id)
            if whereClauseObject:
                query = whereClauseObject.applyWhereObject(query, whereClause)
            data = await session.execute(query)
            data = data.scalars().all()
            if not data:
                res = create_response(status=status.data_is_not_exist())
            else:
                for d in data:
                    await session.delete(d)
                await session.commit()
                res = create_response(status=status.success())
        except Exception as e:
            logger.error(str(e))
            res = create_response(status=status.error(e))
        return res
